#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <time.h>
#include <unistd.h>
#include <sys/io.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/extensions/xf86dga.h>
#include <string.h>


#define VERSION "0.4.2"

#define xiadr(x) outb_p((x),0x37b)	/* define for EPP port acess */
#define xiwr(x) outb_p((x),0x37c)
#define xird() inb(0x37c)
#define OVERSCAN 0x10		//0 // 0x10 to enable overscan

#define OVSCAN 64		/* number of overscan pixels */
#define IMG_X 640
#define SIZE_X (IMG_X + OVSCAN)
#define SIZE_Y 500			/* maximum image size */
#define ALLPIXELS (SIZE_X * SIZE_Y)	/*max number of pixels */

char outfilen[81] = "px";

int expose_time = 200;

struct timeval expend;


int start_line = 0;
int end_line = 499;
int black_frame_count = 0;
int save_mode = 0;
int scale = 4;			// image is divided by scale before displaying

int acq_count = 0;		/* Number of frames read since the progeam was started */

int sns_temp;			/* sensor temperature */

int show_hist = 0;		/* flag to enable histogram window */

int run;

int zoomx = SIZE_X / 2;
int zoomy = SIZE_Y / 2;	// zoom window center coordinates

int trackx = SIZE_X / 2;
int tracky = SIZE_Y / 2;	// track window center coordinates

int tracken = 0;		// flag to enable tracking

int bias;

Display *d;

Window w, wh, wt;

Visual *visual;

unsigned int depth;

GC gc_black, gc_white;

int *raw_picture;

int *big_picture;

int *zero_pixels;

unsigned long *addr;		// image area for main screen
unsigned long *addrz;		// image area for zoom window
unsigned long *trk;		// image area for tracking window

XImage *image;

XImage *zoomimg;

XImage *trkimg;

int mode, logsc;

int logtab[1024];

char *content;

int reader_state;

int global_inhibit;

long long loops_per_ms = 10000;

int make_zoom_win = 0;

int bright = 10;

/* delay routines (usleep is not reliable on all machines) */

void u_delay(int us)
{
	long long i;
	for (i=0; i < us * loops_per_ms / 1000; i++)
		;
}

/* calibrate loops_per_ms */
void calibrate_delay(void)
{
	struct timeval t1, t2;
	int delta;

	loops_per_ms = 1000;
	printf("calibrating delay...\n");
	gettimeofday(&t1, NULL);
	u_delay(1000000);
	gettimeofday(&t2, NULL);
	delta = (t2.tv_usec - t1.tv_usec) + 1000000 * (t2.tv_sec - t1.tv_sec);
	loops_per_ms = 1000000000 / delta;
	printf("%d loops per ms\n", loops_per_ms);
}


// *** Low level parallel port interface routines ***

int i2cd = 0x03;

void setsda(void)
{
	i2cd |= 0x02;
	xiadr(3);
	xiwr(i2cd);
}

void clrsda(void)
{
	i2cd &= 0x01;
	xiadr(3);
	xiwr(i2cd);
}

void setscl(void)
{
	i2cd |= 0x01;
	xiadr(3);
	xiwr(i2cd);
}

void clrscl(void)
{
	i2cd &= 0x02;
	xiadr(3);
	xiwr(i2cd);
}

int rdsda(void)
{
	xiadr(3);
	if (xird() & 0x02)
		return (1);
	else {
		return (0);
	}
}

void i2cwr(unsigned int adr, unsigned int data)
{
	int i;
	clrsda();		/* send start */
	clrscl();

	for (i = 0; i < 7; i++) {	/* send address */
		if (adr & 0x40)
			setsda();
		else
			clrsda();
		adr <<= 1;
		setscl();
		clrscl();
	}

	clrsda();		/* r/-w */
	setscl();
	clrscl();

	setsda();		/* ack input */
	setscl();
	clrscl();

	for (i = 0; i < 8; i++) {	/* send data */
		if (data & 0x80)
			setsda();
		else
			clrsda();
		data <<= 1;
		setscl();
		clrscl();
	}

	setsda();		/* ack input */
	setscl();
	clrscl();

	setscl();		/* stop */
}

unsigned int i2crd(unsigned int adr)
{
	int i;
	unsigned data;

	data = 0;
	clrsda();		/* send start */
	clrscl();

	for (i = 0; i < 7; i++) {	/* send address */
		if (adr & 0x40)
			setsda();
		else
			clrsda();
		adr <<= 1;
		setscl();
		clrscl();
	}

	setsda();		/* r/-w */
	setscl();
	clrscl();

	setsda();		/* ack input */
	setscl();
	clrscl();

	for (i = 0; i < 8; i++) {	/* read data */
		setsda();
		setscl();
		data = data * 2;
		data += rdsda();
		clrscl();
	}

	setsda();		/* ack input */
	setscl();
	clrscl();

	clrsda();
	setscl();		/* stop */
	setsda();

	return (data);
}


void adc_pgm(int addr, int val)
{				/* write to adc control regs */
	int i, d;

	xiadr(0);
	xiwr(4);		/* SLOAD = 1; SDAT = 0; SCLK = 0 */

	d = 0;			/* init holding register */
	xiwr(d);		/* R/W = 0 */
	xiwr(d | 02);
	xiwr(d);		/* pulse SCLK */

	for (i = 0; i < 3; i++) {	/* output address */
		d = (d & 6) | ((addr & 4) == 4 ? 1 : 0);
		xiwr(d);	/* Write address bit */
		xiwr(d | 02);
		xiwr(d);	/* pulse SCLK */
		addr <<= 1;
	}

	for (i = 0; i < 12; i++) {	/* output data */
		d = (d & 6) | ((val & 0x800) == 0x800 ? 1 : 0);
		xiwr(d);	/* Write data bit */
		xiwr(d | 02);
		xiwr(d);	/* pulse SCLK */
		val <<= 1;
	}

	xiwr(4);		/* raise SLOAD */
}

void init_cpix(void)
{				/* init board, incl adc registers */
	int i;

	// the following block resets the port somehow

	outb_p(0x02, 0x37A);	// make program# = 0  
	outb_p(0x00, 0x378);	// make din = 0
	u_delay(1000);
	outb_p(0x06, 0x37A);	// init all strobes
	u_delay(1000);
	outb_p(0x02, 0x37A);	// make program# = 0
	outb_p(0x00, 0x378);	// make din = 0
	u_delay(1000);
	outb_p(0x06, 0x37A);	// init all strobes
	u_delay(1000);
	outb_p(0x04, 0x37A);	// set strobes to default values

/* set epp mode and clear timeout bit */
	outb_p(0x80, 0x77a);
	outb_p(1, 0x379);

	xiadr(4);
	xiwr(0x55);
	printf("testreg reads %x [aa]\n", xird());
	xiadr(4);
	xiwr(0xaa);
	printf("testreg reads %x [55]\n", xird());

	// now begin initialising PLD registers

	xiadr(0);
	xiwr(4);

	xiadr(3);
	xiwr(0x03);

	// following is adc initialisation

	adc_pgm(0, 0x0050);
	adc_pgm(1, 0x0040);
	adc_pgm(2, 0x00);
	adc_pgm(3, 0x00);
	adc_pgm(4, 0x00);
	/* adc_pgm(5, 0x10); */
	adc_pgm(5, 0x1ff);
	adc_pgm(6, 0x00);
	adc_pgm(7, 0x00);

	xiadr(1);

	xiwr(9);
	xiwr(4);

	xiadr(2);
	xiwr(125 - start_line / 4);	// starting line of image divided by 4

	i2cwr(0x2a, 0x01);

	xiadr(1);
	xiwr(4);

	printf("adc programmed \n");
}

struct timeval tv_old;

void zero_timer(void)
{				/* reset the exposure timer */
	struct timeval tv;
	struct timezone tz;

	tz.tz_minuteswest = -180;
	gettimeofday(&tv, &tz);
	tv_old.tv_usec = tv.tv_usec;	/* update variables */
	tv_old.tv_sec = tv.tv_sec;
}

unsigned int get_timer(void)
{				/* read the exposure timer
				   in ms from the last call to zero_timer */
	unsigned int deltat;
	struct timeval tv;
	struct timezone tz;

	tz.tz_minuteswest = -180;
	gettimeofday(&tv, &tz);
	deltat = (tv.tv_sec - tv_old.tv_sec) * 1000 +
	    (tv.tv_usec - tv_old.tv_usec) / 1000;
	return (deltat);
}

void set_expose(void)
{				/* setup an exposure 
				   if it's called soon enough after the last time 
				   sensor was read, do nothing; otherwise, reset the 
				   sensor and reinit time variables */
	int i;

	if ((get_timer() + 10) < expose_time)
		return;

	/* if time is tight, reset the sensor */

	zero_timer();
	fprintf(stderr, "SNSRES ");
	xiadr(1);
	xiwr(0);
	for (i = 0; i < 100; i++);
	xiwr(4);
}

void end_expose(void)
{				/* wait until the exposure time elapses,
				   zero the timer, transfer the data into 
				   sensor memory, and set everything for readout */
	int j, i;

	for (i = 0; i < 5000000; i++) {	/* wait for exposure end */
		if (get_timer() >= expose_time)
			break;
		else if (expose_time - get_timer() > 15)
			u_delay(10000);
	}

	gettimeofday(&expend, NULL);

	i2cwr(0x2a, 0x01);	/* setup for temperature read */

	fprintf(stderr, "Frame: %d, Temp = %d, Exptime = %d ms ",
		acq_count - 1, (sns_temp =
				((j = i2crd(0x2a)) < 127 ? j : j - 256)),
		get_timer());

	zero_timer();

	xiadr(1);		/* start transfer to sensor storage area */

	xiwr(5);
	u_delay(1000);
	xiwr(4);

	xiadr(2);
	xiwr(125 - start_line / 4);

	xiadr(1);
	xiwr(0x06);
	for (j = 0; j < 1000; j++);

	xiadr(3);

	for (j = 0; j < 100; j++) {	/* wait for transfer to complete */
		if ((xird() & 0x04) != 0)
			break;
		u_delay(10000);
	}

	if (j == 100)
		fprintf(stderr, " Timeout in end_expose\n");

	xiadr(1);		/* get ready for readout */
	xiwr(0x04 | OVERSCAN);
}

#define RDLINE_TIMEOUT 1000

void ReadOneLineFromRam(int *destination)
{
	int i, j, x, y, v;

	if (inb_p(0x379) & 0x01) {/* timeout bit */
		outb_p(1, 0x379);
		printf("rdline: reseting epp timeout\n");
	}

	for (j = 0; j < RDLINE_TIMEOUT; j++) {
		xiadr(3);
		if ((xird() & 0x04) != 0)
			break;
		else 
			u_delay(10);
	}

	if (j == RDLINE_TIMEOUT)
		fprintf(stderr, "Timeout in read_line [3] = %x\n", xird());

	xiadr(0);
	//      fprintf(stderr,";" ) ;
	bias = 0;
	xird();
	for (j = 0; j < 20; j++) {	/* discard first dummy pixels */
		y = xird();
		x = xird();
	}
	for (j = 0; j < 23; j++) {	/* discard pixels near the mask */
		y = xird();
		x = xird();
	}
	for (j = 0; j < SIZE_X; j++) {
		y = xird();
		x = xird();
		v = (y * 256 + x) / 4;
		destination[j] = v;
	}

	if (OVERSCAN != 0) {
		for (j = SIZE_X - OVSCAN + 8; j < SIZE_X; j++)	// average overscan for bias
		{
			bias += destination[j];
		}
		bias /= (OVSCAN - 8);
		for (j = 0; j < SIZE_X; j++) {
			if (destination[j] + 16 > bias)
				destination[j] -= (bias - 16);
			else
				destination[j] = 0;
		}
	}

	xird();

	xiadr(1);
	xiwr(0x0c | OVERSCAN);	/* enable new line */
	u_delay(1000);
	xiwr(0x04 | OVERSCAN);	/* enable new line */
}

int ReadImgFile(int *image_array, char *filename)
{
	char lb[81];
	FILE *fp;
	int i, j, k, ef;
	unsigned v;

	ef = 0;
	lb[80] = 0;

	if (strcmp(filename + (strlen(filename) - 5), ".fits") != 0)
		strcat(filename, ".fits");

	fp = fopen(filename, "r");
	if (fp == NULL) {
		fprintf(stderr, "\nCannot read file: %s\n", filename);
		return (1);
	}

	for (i = 0; i < 6; i++)	// at most 6 header blocks
	{
		for (j = 0; j < 36; j++)	// 36 cards per block
		{
			for (k = 0; k < 80; k++)	// chars per card
				lb[k] = fgetc(fp);
			if (strncmp(lb, "END", 3) == 0)
				ef = 1;
		}
		if (ef == 1)
			break;
	}

	if (ef == 0) {
		fprintf(stderr,
			"\nBad FITS format: cannot find END keyword\n");
		return (2);
	}

	printf("\n Reading FITS file: %s\n", filename);

	for (i = 0; i < ALLPIXELS; i++) {
		v = 256 * getc(fp);
		v += getc(fp);
		image_array[i] = v;
		if (i < 50)
			printf(" %d", image_array[i]);
	}
	fclose(fp);
	return (0);
}

int WriteImgFile(int *image_array, char *filename)
{
	char lb[81];
	FILE *fp;
	int i, j, k, v;
	struct tm *t;
	float *dp;
	int jdi;
	double jd;

	lb[80] = 0;

	if (strcmp(filename + strlen(filename) - 5, ".fits") != 0)
		strcat(filename, ".fits");

	fp = fopen(filename, "w");
	if (fp == NULL) {
		fprintf(stderr, "\nCannot open file: %s for writing\n",
			filename);
		return (1);
	}


	t = gmtime(&(expend.tv_sec));
	i = 1900 + t->tm_year;
	j = t->tm_mon + 1;
	k = t->tm_mday;
	jdi = k - 32075 + 1461 * (i + 4800 + (j - 14) / 12) / 4
		+ 367 * (j - 2 - (j - 14) / 12 * 12) / 12
		- 3 * (( i + 4900 + (j - 14) / 12) / 100) / 4;
	jd = jdi - 0.5 + t->tm_hour / 24.0 + t->tm_min / (24.0 * 60.0) 
		+ t->tm_sec / (24.0 * 3600.0) + (expend.tv_usec - expose_time * 500.0)/ 
		(1000000.0 * 24.0 * 3600.0);

	fprintf(fp, "%-8s= %20s / %40s       ", "SIMPLE", "T",
		"Standard FITS format");
	fprintf(fp, "%-8s= %20d / %40s       ", "BITPIX", 16,
		"Bits per pixel");
	fprintf(fp, "%-8s= %20d / %40s       ", "NAXIS", 2, "");
	fprintf(fp, "%-8s= %20d / %40s       ", "NAXIS1", IMG_X, "");
	fprintf(fp, "%-8s= %20d / %40s       ", "NAXIS2",
		end_line - start_line + 1, "");
	fprintf(fp, "%-8s= %20.3f / %40s       ", "EXPTIME",
		0.001 * expose_time, "");
	fprintf(fp, "%-8s= %20.1f / %40s       ", "SNS_TEMP",
		273.0 + sns_temp, "Temperature in K");
	fprintf(fp, "%-8s= %20.6f / %40s       ", "JDATE", jd,
		"");
	fprintf(fp, "%-8s= %20s / %40s       ", "END", "", "");

	for (i = 0; i < 36 - 9; i++)
		fprintf(fp, "%80s", "");

	for (i = start_line; i <= end_line; i++)
		for (j = 0; j < IMG_X; j++) {
			v = image_array[i * SIZE_X + j];
			putc((v >> 8) & 0xff, fp);
			putc((v) & 0xff, fp);
		}

	for (i = 0;
	     i < 2880 - ((IMG_X * (end_line - start_line + 1)) % 2880);
	     i++)		/* pad the end of record */
		putc(' ', fp);
	fclose(fp);
}


void AddToBlackRef(void)
{
	int i;
	fprintf(stderr, " adding to bref ");
	for (i = 0; i < ALLPIXELS; i++)
		zero_pixels[i] += raw_picture[i];
}

void NormSaveBlackRef(void)
{
	int i;
	fprintf(stderr, " normalising bref ");
	if (black_frame_count != 0)
		for (i = 0; i < ALLPIXELS; i++)
			zero_pixels[i] /= black_frame_count;
	WriteImgFile(zero_pixels, outfilen);
}

void SaveFitsImage(void)
{
	char fno[16];
	char nfn[256];
	sprintf(fno, "%d", acq_count - 1);
	strncpy(nfn, outfilen, 240);
	strcat(nfn, fno);
	WriteImgFile(raw_picture, nfn);
	fprintf(stderr, " Saved ");
}

// *** The Image Reading State Machine ***

void Redraw();
void ReadMachine()
{
	static int current = 0, i;
	switch (reader_state) {
	case 0:
		zero_timer();
		reader_state = 1;
		break;
	case 1:
		acq_count++;
		set_expose();
		reader_state = 2;
		break;
	case 2:
		if ((get_timer() > expose_time - 10))
			//|| ((get_timer() > 2000) && acq_count == 1))
		{
			end_expose();
			current = 0;
			reader_state++;
		}
		usleep(10000);
		break;
	case 3:
		for (i = start_line; i <= end_line; i++) {
			ReadOneLineFromRam(raw_picture + i * SIZE_X);
			if (i % 64 == 0)
				fprintf(stderr, ".");
		}
		if ((acq_count > 1)
		    && (acq_count <=
			(black_frame_count + 1))) AddToBlackRef();
		if (acq_count == (black_frame_count + 1)
		    && black_frame_count != 0)
			NormSaveBlackRef();
		if ((black_frame_count == 0) && (acq_count > 1)
		    && outfilen != NULL && save_mode == 1)
			SaveFitsImage();
		reader_state = 1;
		Redraw();
		break;
	default:
		reader_state = 1;
		break;
	}
}

// *** File output functions ***

void SaveRawPicture()
{
	printf("Raw picture file name: ");
	char name[256];
	scanf("%s", &name);
	FILE *fp = fopen(name, "w");
	if (fp) {
		fwrite(big_picture, sizeof(int), ALLPIXELS, fp);
		fclose(fp);
		printf("Done.\n");
	} else
		printf("Couldn't save picture to %s\n", name);
}

void SaveJPEG()
{
	int i;

	printf("JPEG file name: ");
	char name[256], command[256];
	scanf("%s", &name);
	FILE *fp = fopen("gg", "w");
	if (fp) {
		for (i = 0; i < ALLPIXELS; i++) {
			int v = (raw_picture[i] - zero_pixels[i]) / 4;
			if (v < 0)
				v = 0;
			if (v > 255)
				v = 255;
			fputc(v, fp);
			fputc(v, fp);
			fputc(v, fp);
		}
		fclose(fp);
		sprintf(command, "rawtoppm %d %d <gg >jj", SIZE_X, SIZE_Y);
		printf("Executing: %s\n", command);
		system(command);
		sprintf(command, "cjpeg <jj >%s.jpg", name);
		printf("Executing: %s\n", command);
		system(command);
		unlink("gg");
		unlink("jj");
		printf("Done.\n");
	} else
		printf("Couldn't save JPEG to %s\n", name);
}

// *** Some elementary operations ***



void ShowDarkLevels()
{
	int i;
	for (i = 0; i < ALLPIXELS; i++)
		big_picture[i] = zero_pixels[i];
	content = "(pre)set dark image";
}

void ShowRawPicture()
{
	int i;
	for (i = 0; i < ALLPIXELS; i++)
		big_picture[i] = raw_picture[i];
	content = "raw sensor image";
}

void SetBlackLevel()
{
	int i;
	for (i = 0; i < ALLPIXELS; i++)
		zero_pixels[i] = raw_picture[i];
}

void ZeroBlackLevel()
{
	int i;
	for (i = 0; i < ALLPIXELS; i++)
		zero_pixels[i] = 0;
}

void DummyPicture()
{
	int i;
	for (i = 0; i < ALLPIXELS; i++) {
		zero_pixels[i] = (rand() & 31);
		raw_picture[i] = (rand() & 31);
	}
	for (i = 0; i < 5000; i++) {
		int x = (rand() & 2047);
		int y = (rand() & (SIZE_Y - 1));
		int l = (rand() & 65535);
		if (x >= SIZE_X)
			x -= SIZE_X;
		raw_picture[y * SIZE_X + x] = l;
	}
}


/* scale, dark-substract and transfeer the sensor image to
 * the format X will want*/
void transfer_picture_16(int *p)
{

	int xz, yz, x, y;
	int tpkx, tpky;
	int v;
	char nm[256];
	double val, zavg, zpk, sx, sx2, zvar;
	double tavg, tpk, tvar;
	unsigned short *a;

	a = (unsigned short *)addr;

	for (y = 0; y < SIZE_Y; y++)	// transfer to the large window
		for (x = 0; x < SIZE_X; x++) {
			v = p[y * SIZE_X + x];
			if (v > zero_pixels[y * SIZE_X + x] - 32)
				v = v + 32 - zero_pixels[y * SIZE_X + x];
			else
				v = 0;
			big_picture[y * SIZE_X + x] = v;
			/* if(logsc) v=logtab[v] ; */
			v = v / (scale) + bright;
			if (v > 255)
				v = 255;
			v /= 8;
			a[(y) * SIZE_X + x] = (v) * 0x841;
		}

	sx = 0.0;
	sx2 = 0.0;
	zpk = 0.0;		// compute statistics of the zoom window
	for (x = 0; x < 64; x++)
		for (y = 0; y < 64; y++) {
			val =
			    1.0 * big_picture[zoomx + x - 32 +
					      SIZE_X * (zoomy + y - 32)];
			if (val > zpk)
				zpk = val;
			sx += val;
			sx2 += val * val;
		}
	zavg = sx / 4096.0;
	zvar = sx2 / 4096.0 - zavg * zavg;
//	sprintf(nm, "Z a=%1.0f p=%1.0f s=%1.0f", zavg, zpk, sqrt(zvar));
//	XStoreName(d, wh, nm);
}

void TransferPicture(int *p)
{
	int xz, yz, x, y, i;
	int tpkx, tpky;
	int v;
	char nm[256];
	double val, zavg, zpk, sx, sx2, zvar;
	double tavg, tpk, tvar;

	if (depth == 16) {
		transfer_picture_16(p);
		return;
	}

	/*  for(xz = 0; xz < 256 * 256; xz++)
	   addrz[xz] = 0; */
	for (y = 0; y < SIZE_Y; y++)	// transfer to the large window
		for (x = 0; x < SIZE_X; x++) {
			v = p[y * SIZE_X + x];
			if (v > zero_pixels[y * SIZE_X + x] - 32)
				v = v + 32 - zero_pixels[y * SIZE_X + x];
			else
				v = 0;
			big_picture[y * SIZE_X + x] = v;
			/* if(logsc) v=logtab[v] ; */
			v = v / (scale) + bright;
			if (v > 255)
				v = 255;
			addr[(y) * SIZE_X + x] = (v) * 0x00010101;

		}
	if (make_zoom_win) {
		sx = 0.0;
		sx2 = 0.0;
		zpk = 0.0;		// compute statistics of the zoom window
		for (x = 0; x < 64; x++)
			for (y = 0; y < 64; y++) {
				val =
					1.0 * big_picture[zoomx + x - 32 +
							  SIZE_X * (zoomy + y - 32)];
				if (val > zpk)
					zpk = val;
				sx += val;
				sx2 += val * val;
			}
		zavg = sx / 4096.0;
		zvar = sx2 / 4096.0 - zavg * zavg;
		sprintf(nm, "Z avg=%1.0f pk=%1.0f sd=%1.0f", zavg, zpk, sqrt(zvar));
		XStoreName(d, wh, nm);
		
		for (x = 0; x < 64; x++)	// scale and transfer zoom window
			for (y = 0; y < 64; y++) {
				val =
					1.0 * big_picture[zoomx + x - 32 +
							  SIZE_X * (zoomy + y - 32)];
				val -= (zavg - 3.0 * sqrt(zvar));
				if (val < 0.0)
					val = 0.0;
				double s;
			s = (zpk - (zavg - 3.0 * sqrt(zvar)));
			if (s < 100)
				s = 100;
			for (yz = 0; yz < 4; yz++)
				for (xz = 0; xz < 4; xz++) {
					addrz[x * 4 + xz +
					      256 * (y * 4 + yz)] =
						((int) floor(val * 255.0 / s))
						* 0x00010101;
				}
			if (x == 0 || x == 63 || y == 0 || y == 63)
				addr[zoomx + x - 32 +
				     SIZE_X * (zoomy + y - 32)] |=
					0x00ff0000;
			}
		for (yz = 0; yz < 256; yz++) {
			addrz[yz * 256 + 111] |= 0x00ff0000;
			addrz[yz * 256 + 144] |= 0x00ff0000;
			addrz[111 * 256 + yz] |= 0x00ff0000;
			addrz[144 * 256 + yz] |= 0x00ff0000;
		}
	}
	// now process the track window

	sx = 0.0;
	sx2 = 0.0;
	tpk = 0.0;		// compute statistics of the track window
	tpkx = tpky = 0;
	for (x = 16; x < 48; x++)
		for (y = 16; y < 48; y++) {
			val =
			    1.0 * big_picture[trackx + x - 32 +
					      SIZE_X * (tracky + y - 32)];
			if (val > tpk) {
				tpk = val;
				tpkx = trackx + x - 32;
				tpky = tracky + y - 32;
			}
			sx += val;
			sx2 += val * val;
		}
	tavg = sx / 1024.0;
	tvar = sx2 / 1024.0 - tavg * tavg;

	// now compute the tracking parameters

	int r[4] = { 0, 0, 0, 0 };
	int rt = 0;

	for (i = 1; i < 6; i++) {
		if ((1.0 * big_picture[tpkx + i + SIZE_X * tpky]) >
		    tpk / 2) r[0] = i;
		if ((1.0 * big_picture[tpkx - i + SIZE_X * tpky]) >
		    tpk / 2) r[1] = i;
		if ((1.0 * big_picture[tpkx + SIZE_X * (tpky + i)]) >
		    tpk / 2) r[2] = i;
		if ((1.0 * big_picture[tpkx + SIZE_X * (tpky - i)]) >
		    tpk / 2) r[3] = i;
	}
	rt = (r[0] + r[1] + r[2] + r[3]) / 4 + 6;
	double trkofsx = 0.0;
	double trkofsy = 0.0;
	double sum = 0.0;

	for (x = -7; x < 8; x++)
		for (y = -7; y < 8; y++) {
			if ((x * x + y * y) <= rt * rt) {
				sum +=
				    (1.0 *
				     big_picture[tpkx + x +
						 SIZE_X * (tpky + y)]);
				trkofsx +=
				    x * (1.0 *
					 big_picture[tpkx + x +
						     SIZE_X * (tpky + y)]);
				trkofsy +=
				    y * (1.0 *
					 big_picture[tpkx + x +
						     SIZE_X * (tpky + y)]);
			}
		}
}

// *** A powerful (?!) event processor ***

void ShowHistogram(int *p)
{
	int levels[256], i;
	float npix, sumx = 0.0, sumx2 = 0.0, avg, avg2, sigma;
	for (i = 0; i < 256; i++)
		levels[i] = 0;
	for (i = 8 * SIZE_X; i < ALLPIXELS - 64 * SIZE_X; i++) {
		if ((p[i] < 16384) && (p[i] >= 0)) {
			levels[p[i] / 256]++;
			sumx += 1.0 * p[i];
		}
	}
	npix = (ALLPIXELS - 72.0 * SIZE_X);
	avg = sumx / npix;
	for (i = 8 * SIZE_X; i < ALLPIXELS - 64 * SIZE_X; i++) {
		sumx2 += ((1.0 * p[i]) - avg) * ((1.0 * p[i]) - avg);
	}
	sigma = sqrt(sumx2 / npix);

	fprintf(stderr, "\navg = %8.2f   sigma = %8.2f  bias = %d\n", avg,
		sigma, bias);

	if (show_hist == 1) {
		int max = 0, div;
		for (i = 0; i < 256; i++)
			if (levels[i] > max)
				max = levels[i];
		div = max / 256;

		for (i = 0; i < 256; i++) {
			int v = levels[i] / div;
			XDrawLine(d, wh, gc_white, i, 256, i, 256 - v);
			XDrawLine(d, wh, gc_black, i, 256 - v, i, 0);
		}
	}
}

void Redraw()
{
	TransferPicture(raw_picture);
	ShowHistogram(big_picture);
	XPutImage(d, w, gc_black, image, 0, 0, 0, 0, SIZE_X, SIZE_Y);
	if (zoomimg)
		XPutImage(d, wh, gc_black, zoomimg, 0, 0, 0, 0, 256, 256);
//	XPutImage(d, wt, gc_black, trkimg, 0, 0, 0, 0, 256, 256);
	char name[256];
}

void ProcessEvent(XEvent * evt)
{
	int recalc = 0, redraw = 0;
	KeySym key;
	switch (evt->type) {
	case KeyPress:
		key = XLookupKeysym(&evt->xkey, 0);
		redraw = 1;
		switch (key) {
		case XK_q:
			run = 0;
			break;
		case XK_space:
			printf("rstat %d exptime %d acqcount %d bkref %d\n", 
			       reader_state, get_timer(), acq_count, black_frame_count);
			recalc = 1;
			break;
		case XK_Escape:
			init_cpix();
			break;
		case XK_b:
			SetBlackLevel();
			break;
		case XK_z:
			ZeroBlackLevel();
			break;
		case '0':
			scale = 1;
			Redraw();
			break;
		case '1':
			scale = 2;
			Redraw();
			break;
		case '2':
			scale = 4;
			Redraw();
			break;
		case '3':
			scale = 8;
			Redraw();
			break;
		case '4':
			scale = 16;
			Redraw();
			break;
		case '5':
			scale = 32;
			Redraw();
			break;
		case '6':
			scale = 64;
			Redraw();
			break;
		case '7':
			scale = 128;
			Redraw();
			break;
		case '8':
			scale = 256;
			Redraw();
			break;
		case XK_Return:
			reader_state = 1;
			break;
		case XK_h:
			global_inhibit = 1;
			break;
		case XK_e:
			if (bright < 122)
				bright += 4;
			break;
		case XK_d:
			if (bright >= 4)
				bright -= 4;
			break;
		case XK_c:	// zet zoom center command
			zoomx = evt->xkey.x;
			zoomy = evt->xkey.y;
			if (zoomx < 32)
				zoomx = 32;
			if (zoomx > SIZE_X - 32)
				zoomx = SIZE_X - 32;
			if (zoomy < 32)
				zoomy = 32;
			if (zoomy > SIZE_Y - 32)
				zoomy = SIZE_Y - 32;
			Redraw();
			break;
		case XK_y:	// enable tracking command
			tracken = 1;
			break;
		case XK_s:	// stop tracking command
			tracken = 0;
			break;
		}
		break;
	case Expose:
		if (evt->xexpose.count == 0)
			redraw = 1;
		break;
	}
	if (recalc) {
		TransferPicture(big_picture);
		redraw = 1;
	}
	if (redraw) {
		XPutImage(d, w, gc_black, image, 0, 0, 0, 0, SIZE_X,
			  SIZE_Y);
	}
}

int main(int argc, char **argv)
{
	int ap, i;
	FILE *fp;
	// Allocate images

	big_picture = malloc(ALLPIXELS * sizeof(int));
	raw_picture = malloc(ALLPIXELS * sizeof(int));
	zero_pixels = malloc(ALLPIXELS * sizeof(int));
	addr = malloc(ALLPIXELS * sizeof(int));
	addrz = malloc(256 * 256 * sizeof(int));
	trk = malloc(256 * 256 * sizeof(int));;

	// Process command-line parameters

	if (argc == 1) {
		printf("\n cpx version "VERSION"\n");
		printf("\n Usage: cpx option ... filename");
		printf("\n Options:");
		printf("\n -e <expose time> (in milliseconds)");
		printf("\n -a <start line> <end line>");
		printf("\n -b <nr of black frames>");
		printf("\n -r <black ref file>");
		printf("\n -z -- show zoom window (24/32 bpp only)");
		printf("\n -s -- save images");
		printf("\n");
		printf("\n In image window, press '1'-'8' to adjust contrast");
		printf("\n 'C' to set zoom center, 'B' to set black ref,\n");
		printf("\n 'Z' to clear black ref, 'D' to make image darker,\n");
		printf("\n 'B' to make it brighter, 'Q' to quit\n");
		exit(0);
	}

	for (ap = 1; ap < argc; ap++) {
		if (argv[ap][0] == '-')
			switch (argv[ap][1]) {
			case 'h':	// print help
				show_hist = 1;
				break;

			case 'e':	// set expose time
				expose_time = atoi(argv[++ap]);
				if (expose_time < 10)
					expose_time = 10;
				if (expose_time > 100000)
					expose_time = 100000;
				printf("\nExpose set to %d ms\n",
				       expose_time);
				break;

			case 'a':	// select area on screen
				start_line = atoi(argv[++ap]);
				end_line = atoi(argv[++ap]);
				start_line &= 0xfffc;
				end_line &= 0xfffc;
				if (start_line < 0)
					start_line = 0;
				if (start_line >= SIZE_Y - 4)
					start_line = SIZE_Y - 5;

				if (end_line < start_line + 4)
					end_line = start_line + 4;
				if (end_line >= SIZE_Y)
					end_line = SIZE_Y - 1;
				break;

			case 'b':	// enable taking of a number of black frames
				black_frame_count = atoi(argv[++ap]);
				if (black_frame_count < 1)
					black_frame_count = 1;
				if (black_frame_count > 100)
					black_frame_count = 100;

				for (i = 0; i < ALLPIXELS; i++)	// init black reference buffer
					zero_pixels[i] = 0;

				break;

			case 'r':	// load black reference file
				ReadImgFile(zero_pixels, argv[++ap]);
				break;

			case 'z':	// create zoom window
				make_zoom_win = 1;
				break;

			case 's':	// select save mode
				save_mode = 1;
				break;
		} else {
			if (argv[ap] != NULL)
				strncpy(outfilen, argv[ap], 64);
		}
	}

	calibrate_delay();
	acq_count = 0;

	global_inhibit = 0;
//	printf("ioperm1 %d\n", ioperm(0x378, 8, 1));
	if (iopl(3)) {
		printf("Cannot get permissions to use i/o ports (must be suid root)\n");
		exit(1);
	}
	d = XOpenDisplay("");
	if (d) {
		int scr = DefaultScreen(d);
		Colormap cmap = DefaultColormap(d, scr);
		visual = DefaultVisual(d, scr);
		depth = DefaultDepth(d, scr);
		printf("depth is %d\n", depth);

		if (depth < 16) {
			fprintf(stderr, "need at least 16bpp, aborting\n");
			exit(1);
		}
		if (make_zoom_win && depth == 16) {
			fprintf(stderr, "zoom window is only supported at 24bpp and up\n");
			make_zoom_win = 0;
		}
		unsigned long white = WhitePixel(d, scr), black =
		    BlackPixel(d, scr);
		w =
		    XCreateSimpleWindow(d, DefaultRootWindow(d), 0, 0,
					SIZE_X, SIZE_Y, 1, white, black);
		wh = XCreateSimpleWindow(d, DefaultRootWindow(d), 0, 0, 256,
					 256, 1, white, black);
//		wt =
//		    XCreateSimpleWindow(d, DefaultRootWindow(d), 0, 0, 256,
//					256, 1, white, black);
		gc_black = XCreateGC(d, w, 0, 0);
		XSetForeground(d, gc_black, black);
		gc_white = XCreateGC(d, w, 0, 0);
		XSetForeground(d, gc_white, white);
		XSelectInput(d, w,
			     KeyPressMask | ExposureMask |
			     ButtonPressMask);
		XSelectInput(d, wh, KeyPressMask);
//		XSelectInput(d, wt, KeyPressMask);
		XStoreName(d, w, "CPX "VERSION);
		XMapRaised(d, w);
		if (make_zoom_win) {
			XStoreName(d, wh, "CPX Zoom Window");
			XMapRaised(d, wh);
		}
//		XStoreName(d, wt, "Tracking");
//		XMapRaised(d, wt);
	} else {
		fprintf(stderr, "No X-display available\n");
		return 1;
	}

	for (i = 0; i < 1024; i++)
		logtab[i] = floor((log10((i) / 40.0 + 1.0) / log10(1025.0 / 40.0) *
				   1023.0));

	TransferPicture(big_picture);

	image = XCreateImage(d, visual, depth, ZPixmap, 0, (char *) addr,
			     SIZE_X, SIZE_Y, 32, 0);
	XInitImage(image);

	if (make_zoom_win) {
		if (depth > 16) {
			zoomimg = XCreateImage(d, visual, depth, ZPixmap, 0, 
					       (char *) addrz, 256, 256, 32, 0);
			XInitImage(zoomimg);
		}
	}

	mode = 0;
	logsc = 0;
	init_cpix();

	for (run = 1; run;) {
		while (run && (	/*reader_state==0 || */
				      XPending(d) > 0)) {
			XEvent evt;
			XNextEvent(d, &evt);
			ProcessEvent(&evt);
		}
		ReadMachine();
	}

	XFree(image);
	free(addr);
	if (zoomimg)
		XFree(zoomimg);
	free(addrz);
//	XFree(trkimg);
//	delete trk;
	XCloseDisplay(d);
	return 0;
}
